from typing import Any, Callable, Type, Dict, Tuple, Set, List, Union
from collections import deque
from loguru import logger
import inspect
import traceback
import types
from pydantic import BaseModel
from typing_extensions import get_origin, get_args

def is_type(value, dst_type):
    origin = get_origin(dst_type)
    if origin is None:
        return isinstance(value, dst_type)

    if origin is Union or origin is types.UnionType:
        args = get_args(dst_type)
        return any(is_type(value, arg) for arg in args)

    if not isinstance(value, origin):
        return False

    args = get_args(dst_type)
    if not args:
        return True
    
    if origin is list:
        elem_type = args[0]
        return all(is_type(item, elem_type) for item in value)
    elif origin is dict:
        key_type, value_type = args
        return all(
            is_type(k, key_type) and is_type(v, value_type)
            for k, v in value.items()
        )
    
    return True

class TypeConverter:
    def __init__(self):
        self._registry: Dict[Tuple[Type, Type], Callable[..., Any]] = {}

    def register(self, src_type: Type, dst_type: Type, func: Callable[..., Any]):
        self._registry[(src_type, dst_type)] = func

    def can_convert(self, src_type: Type, dst_type: Type) -> bool:
        return self._find_path(src_type, dst_type) is not None

    def convert(self, value: Any, dst_type: Type, **kwargs) -> Any:
        if value is None:
            return None
        
        if dst_type == Any:
            return value

        src_type = type(value)

        if is_type(value, dst_type):
            return value

        if (src_type, dst_type) in self._registry:
            return self._call_func(self._registry[(src_type, dst_type)], value, dst_type=dst_type, **kwargs)

        origin = getattr(dst_type, "__origin__", None)
        if origin in (tuple, list) or dst_type in (tuple, list):
            logger.warning(f"Converting to {dst_type} may not be supported. Returning value as-is.")
            return value

        try:
            return dst_type(value)
        except Exception:
            pass

        if issubclass(dst_type, BaseModel) and isinstance(value, dict):
            return dst_type(**value)

        path = self._find_path(src_type, dst_type)
        if not path:
            raise TypeError(f"No conversion path found from {src_type.__name__} to {dst_type.__name__}.")

        result = value
        for i in range(len(path) - 1):
            func = self._registry[(path[i], path[i + 1])]
            result = self._call_func(func, result, dst_type=dst_type, **kwargs)
        return result

    def _call_func(self, func: Callable[..., Any], value: Any, **kwargs) -> Any:
        sig = inspect.signature(func)
        if len(sig.parameters) == 1:
            return func(value)
        else:
            return func(value, **kwargs)

    def _find_path(self, src_type: Type, dst_type: Type) -> List[Type] | None:
        print('src_type', src_type, 'dst_type', dst_type)
        if src_type == dst_type:
            return [src_type]

        graph: Dict[Type, Set[Type]] = {}
        for (s, d) in self._registry.keys():
            graph.setdefault(s, set()).add(d)

        queue = deque([[src_type]])
        visited = {src_type}

        while queue:
            path = queue.popleft()
            current = path[-1]
            for neighbor in graph.get(current, []):
                if neighbor in visited:
                    continue
                new_path = path + [neighbor]
                if neighbor == dst_type:
                    return new_path
                queue.append(new_path)
                visited.add(neighbor)
        return None


from __future__ import annotations
from typing import Any
from service_forge.workflow.node import Node
from service_forge.workflow.port import Port

class AnyGateNode(Node):
    DEFAULT_INPUT_PORTS = [
        Port("input", Any),
    ]

    DEFAULT_OUTPUT_PORTS = [
        Port("output", Any),
    ]

    def __init__(self, name: str):
        super().__init__(name)

    def _run(self, input: Any) -> Any:
        self.activate_output_edges('output', input)


import asyncio
from typing import AsyncIterator
from service_forge.workflow.node import Node
from service_forge.workflow.port import Port
from model.http.test_sse_model import TestSSEModel
from a2a.server.agent_execution.context import RequestContext

class TestTraceBNode(Node):
    DEFAULT_INPUT_PORTS = [
        Port("input", str)
    ]

    DEFAULT_OUTPUT_PORTS = [
        Port("result", str)
    ]

    def __init__(self, name: str):
        super().__init__(name)

    async def _run(self, input: str) -> None:
        print(f" + TestTraceBNode: {input} {self.get_trace_id()} {self.get_span_id()}")
        self.activate_output_edges('result', f"Input[{input}]")


def get_metadata_file_name(name: str, version: str) -> str:
    return "sf-meta.yaml"

def get_service_name(name: str, version: str) -> str:
    return f"sf-{name}-{version.replace('.', '-')}v"

def get_service_url_name(name: str, version: str) -> str:
    return f"{name}-{version.replace('.', '-')}"

def get_url(entry_url: str, service_name: str = "", version: str = "", path: str = "", subdomain: str = "") -> str:
    if not entry_url:
        raise ValueError("entry_config.url is not set")

    url_parts = entry_url.split("://", 1)
    if len(url_parts) != 2:
        raise ValueError(f"Invalid entry_url format: {entry_url}")

    protocol = url_parts[0]
    rest = url_parts[1]

    parts = rest.split("/", 1)
    domain = parts[0]
    base_path = parts[1] if len(parts) > 1 else ""

    if subdomain:
        domain = f"{subdomain}.{domain}"

    url_path_parts = []
    
    if service_name and service_name == 'service-center':
        url_path_parts.append('service_center')
    
    if base_path and subdomain == "":
        url_path_parts.append(base_path)

    if service_name and version:
        if service_name != 'entry' and service_name != 'service-center':
            service_url_name = get_service_url_name(service_name, version)
            url_path_parts.append(service_url_name)

    if path:
        path = path.lstrip("/")
        url_path_parts.append(path)

    final_url = f"{protocol}://{domain}"
    if url_path_parts:
        final_url += "/" + "/".join(url_path_parts)

    return final_url


from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from service_forge.workflow.workflow import Workflow

class WorkflowValidator:
    def __init__(self, workflow: Workflow) -> None:
        self.workflow = workflow

    def validate(self) -> None:
        pass


from typing import Any
from loguru import logger
from pydantic import BaseModel
from service_forge.workflow.node import Node
from service_forge.workflow.port import Port

class DatabaseCastNode(Node):
    DEFAULT_INPUT_PORTS = [
        Port("object", Any),
        Port("cast_type", type),
    ]

    DEFAULT_OUTPUT_PORTS = [
        Port("object", Any),
    ]

    def __init__(self, name: str):
        super().__init__(name)

    async def _run(self, object: Any, cast_type: type) -> None:
        if not isinstance(cast_type, type):
            raise ValueError(f"cast_type must be a class type, got {type(cast_type)}")
        
        if not (hasattr(cast_type, '__table__') or hasattr(cast_type, '__mapper__')):
            raise ValueError(f"cast_type must be a SQLAlchemy model class (inherited from declarative_base()), got {cast_type}")
        
        if isinstance(object, cast_type):
            logger.debug(f"Data is already an instance of {cast_type.__name__}")
            self.activate_output_edges(self.get_output_port_by_name('object'), object)
            return
        
        if isinstance(object, BaseModel):
            object_dict = object.model_dump()
        elif isinstance(object, dict):
            object_dict = object
        else:
            raise ValueError(f"object must be a dict or BaseModel instance, got {type(object)}")
        
        try:
            instance = cast_type(**object_dict)
            logger.debug(f"✓ Cast data to {cast_type.__name__} instance")
            self.activate_output_edges(self.get_output_port_by_name('object'), instance)
        except Exception as e:
            logger.error(f"✗ Failed to cast data to {cast_type.__name__}: {e}")
            raise ValueError(f"Failed to cast data to {cast_type.__name__}: {e}")


from service_forge.sft.config.injector_default_files import DEFAULT_TRAEFIK_INGRESS_YAML
from service_forge.sft.config.injector_default_files import DEFAULT_AUTH_MIDDLEWARE_YAML
from service_forge.sft.config.injector_default_files import DEFAULT_NOAUTH_WORKFLOW_YAML

def test_injector_files():
    result = DEFAULT_TRAEFIK_INGRESS_YAML.format(
        name="test",
        version="1.0.0",
        namespace="test",
        auth_middleware_yaml=DEFAULT_AUTH_MIDDLEWARE_YAML.format(namespace="test")
    ) + DEFAULT_NOAUTH_WORKFLOW_YAML.format(
        name="test",
        version="1.0.0",
        namespace="test",
        path="/docs".lstrip("/"),
        method="GET",
    )
    print(result)

if __name__ == "__main__":
    test_injector_files()


from pathlib import Path

from starlette.staticfiles import StaticFiles

def get_studio_static_files() -> StaticFiles:
    frontend_dist_path = Path(__file__).parent.parent / "frontend" / "static"
    return StaticFiles(directory=str(frontend_dist_path), html=True)

studio_static_files = get_studio_static_files()



import os
import traceback
import shutil
import subprocess
import tarfile
import tempfile
import requests
import yaml
from pathlib import Path

import typer
from omegaconf import OmegaConf
from service_forge.sft.util.logger import log_error, log_info, log_success, log_warning
from service_forge.sft.file.sft_file_manager import sft_file_manager
from service_forge.sft.config.sft_config import sft_config
from service_forge.sft.util.assert_util import assert_file_exists, assert_dir_exists
from service_forge.sft.config.sf_metadata import load_metadata
from service_forge.sft.kubernetes.kubernetes_manager import KubernetesManager
from service_forge.sft.config.injector import Injector
from service_forge.sft.util.name_util import get_service_name

def _extract_tar_file(tar_file: Path, temp_path: Path) -> None:
    log_info(f"Extracting tar file to: {temp_path}")

    try:
        with tarfile.open(tar_file, 'r') as tar:
            tar.extractall(temp_path)
    except Exception as e:
        log_error(f"Failed to extract tar file: {e}")
        raise typer.Exit(1)

    log_success("Tar file extracted successfully")

def _build_docker_image(project_dir: Path, name: str, version: str) -> None:
    image_name = f"sf-{name}:{version}"
    full_image_name = sft_config.registry_address + "/" + image_name
    log_info(f"Building Docker image: {image_name}")
    try:
        # build docker image
        build_result = subprocess.run(
            ["docker", "build", "-t", full_image_name, str(project_dir)],
            capture_output=True,
            text=True,
            check=True
        )
        log_success(f"Docker image built successfully: {image_name}")
        if build_result.stdout:
            log_info(build_result.stdout)

        # push docker image to registry
        log_info(f"Pushing Docker image to registry: {full_image_name}")
        push_result = subprocess.run(
            ["docker", "push", full_image_name],
            capture_output=True,
            text=True,
            check=True
        )
        log_success(f"Docker image pushed successfully: {full_image_name}")
        if push_result.stdout:
            log_info(push_result.stdout)

        # Clean up dangling images (tagged as "none") for this registry
        registry_prefix = sft_config.registry_address
        clear_cmd = f"docker images | grep '^{registry_prefix}/nexthci/sf-' | grep none | awk '{{print $3}}' | xargs -r docker rmi"
        clear_result = subprocess.run(
            clear_cmd,
            shell=True,
            capture_output=True,
            text=True,
            check=False  # Don't fail if no images to clean
        )
        if clear_result.stdout:
            log_info(clear_result.stdout)

    except subprocess.CalledProcessError as e:
        log_error(traceback.format_exc())
        log_error(f"Docker operation failed: {e}")
        if e.stderr:
            log_error(e.stderr)
        raise typer.Exit(1)
    except FileNotFoundError:
        log_error(traceback.format_exc())
        log_error("Docker command not found. Please install Docker.")
        raise typer.Exit(1)

def _apply_k8s_deployment(deployment_yaml: Path, ingress_yaml: Path, name: str, version: str) -> None:
    log_info("Applying k8s deployment...")

    try:
        k8s_manager = KubernetesManager()
        k8s_manager.delete_service(sft_config.k8s_namespace, get_service_name(name, version), force=True)
        k8s_manager.apply_deployment_yaml(deployment_yaml, sft_config.k8s_namespace)
        k8s_manager.apply_deployment_yaml(ingress_yaml, sft_config.k8s_namespace)
        log_success("K8s deployment applied successfully")
    except Exception as e:
        log_error(traceback.format_exc())
        log_error(f"K8s deployment failed: {e}")
        raise typer.Exit(1)

    log_success(f"Deployment process completed for {name}:{version}")

def _inject_config(project_dir: Path) -> None:
    injector = Injector(project_dir)
    injector.inject()

def _send_config_to_api(project_dir: Path, name: str, version: str) -> None:
    """
    部署成功后向API发送配置文件
    """
    # 检查是否配置了notify API URL
    notify_url = sft_config.deploy_notify_api_url
    if not notify_url or notify_url.strip() == "":
        log_info("No deploy_notify_api_url configured, skipping config notification")
        return

    try:
        # 读取service配置文件
        metadata_path = project_dir / "sf-meta.yaml"
        metadata = load_metadata(metadata_path)

        service_config_path = project_dir / metadata.service_config
        if not service_config_path.exists():
            log_warning(f"Service config file not found: {service_config_path}")
            return

        # 读取service配置
        with open(service_config_path, 'r', encoding='utf-8') as f:
            service_config = yaml.safe_load(f)

        # 读取所有workflow配置
        workflow_configs = []
        if 'workflows' in service_config:
            for workflow_path in service_config['workflows']:
                full_workflow_path = (service_config_path.parent / workflow_path).resolve()
                if full_workflow_path.exists():
                    with open(full_workflow_path, 'r', encoding='utf-8') as f:
                        workflow_config = yaml.safe_load(f)
                        workflow_configs.append({
                            'path': workflow_path,
                            'config': workflow_config
                        })

        # 构造发送的数据
        payload = {
            'name': name,
            'version': version,
            'description': metadata.description,
            'service_config': service_config,
            'workflow_configs': workflow_configs
        }

        # 发送POST请求
        log_info(f"Sending config to API: {notify_url}")
        response = requests.post(
            notify_url,
            json=payload,
            headers={'Content-Type': 'application/json'},
            timeout=10
        )

        if response.status_code == 200:
            log_success(f"Successfully sent config to API")
        else:
            log_warning(f"API returned status code: {response.status_code}")
            log_warning(f"Response: {response.text}")

    except Exception as e:
        log_warning(f"Failed to send config to API: {str(e)}")
        log_info("Deployment will continue despite notification failure")


def deploy_service(name: str, version: str) -> None:
    tar_file = sft_file_manager.tar_path / f"sf_{name}_{version}.tar"

    assert_file_exists(tar_file)

    temp_parent = os.path.join(tempfile.gettempdir(), "sft")
    os.makedirs(temp_parent, exist_ok=True)

    with tempfile.TemporaryDirectory(prefix=f"deploy_{name}_{version}", dir=temp_parent) as temp_dir:
        temp_path = Path(temp_dir)

        _extract_tar_file(tar_file, temp_path)

        project_dir = temp_path / f"{name}_{version}"

        _inject_config(project_dir)

        dockerfile_path = project_dir / "Dockerfile"
        metadata_path = project_dir / "sf-meta.yaml"
        deployment_yaml = project_dir / "deployment.yaml"
        ingress_yaml = project_dir / "ingress.yaml"

        assert_dir_exists(project_dir)
        assert_file_exists(dockerfile_path)
        assert_file_exists(metadata_path)
        assert_file_exists(deployment_yaml)
        assert_file_exists(ingress_yaml)

        try:
            meta_data = load_metadata(metadata_path)
        except Exception as e:
            log_error(f"Failed to read sf-meta.yaml: {e}")
            raise typer.Exit(1)

        _build_docker_image(project_dir, meta_data.name, meta_data.version)
        # TODO: create new user in mongodb and redis
        _apply_k8s_deployment(deployment_yaml, ingress_yaml, meta_data.name, meta_data.version)

        # 部署成功后发送配置文件到API
        _send_config_to_api(project_dir, meta_data.name, meta_data.version)



import httpx
from loguru import logger
from fastapi import HTTPException
from enum import Enum
from typing import Any
from service_forge.workflow.node import Node
from service_forge.workflow.port import Port

class CallServiceNodeMethod(Enum):
    GET = "GET"
    POST = "POST"
    PUT = "PUT"
    DELETE = "DELETE"

class CallServiceNode(Node):
    DEFAULT_INPUT_PORTS = [
        Port("service_name", str),
        Port("version", str),
        Port("path", str),
        Port("path_params", dict),
        Port("method", CallServiceNodeMethod),
        Port("data", dict),
        Port("response_type", type),
        Port("user_id", str),
        Port("token", str),
    ]

    DEFAULT_OUTPUT_PORTS = [
        Port("response", Any),
        Port("status_code", int),
    ]

    def __init__(self, name: str):
        super().__init__(name)

    async def _run(
        self,
        service_name: str,
        version: str,
        path: str,
        path_params: dict,
        method: CallServiceNodeMethod,
        data: dict,
        response_type: type,
        user_id: str,
        token: str,
    ) -> Any:
        url = self.get_url(
            service_name=service_name,
            version=version,
            path=path,
            path_params=path_params,
        )
        logger.debug(f"CallServiceNode._run: url: {url} method: {method} data: {data} user_id: {user_id} token: {token}")
        headers = {
            "X-User-ID": user_id,
            "X-Auth-Token": token,
            "Authorization": f"Bearer {token}",
        }
        async with httpx.AsyncClient() as client:
            response = await client.request(
                method.value,
                url,
                json=data,
                headers=headers,
            )

            self.activate_output_edges('status_code', response.status_code)
            if response.status_code == 200:
                self.activate_output_edges('response', response.json())
            else:
                raise HTTPException(status_code=response.status_code, detail=response.json())



import re
import yaml
import subprocess
from pathlib import Path
from service_forge.sft.util.logger import log_info, log_error
from service_forge.sft.config.injector_default_files import *
from service_forge.sft.config.sf_metadata import load_metadata, save_metadata
from service_forge.sft.config.sft_config import sft_config
from service_forge.service_config import ServiceConfig, ServiceFeedbackConfig, SignozConfig, TraceConfig, EntryConfig, LLMConfig
from service_forge.workflow.workflow_config import WorkflowConfig
from service_forge.sft.util.name_util import get_service_name
from service_forge.sft.util.yaml_utils import load_sf_metadata_as_json

class Injector:
    def __init__(self, project_dir: Path):
        self.project_dir = project_dir
        self.deployment_yaml_path = project_dir / "deployment.yaml"
        self.metadata_path = project_dir / "sf-meta.yaml"
        self.ingress_yaml_path = project_dir / "ingress.yaml"
        self.dockerfile_path = project_dir / "Dockerfile"
        self.pyproject_toml_path = project_dir / "pyproject.toml"
        self.start_sh_path = project_dir / "start.sh"
        self.metadata = load_metadata(self.metadata_path)
        self.metadata.mode = "release"
        self.name = self.metadata.name
        self.version = self.metadata.version
        self.namespace = sft_config.k8s_namespace

        try:
            self.sf_metadata_string = load_sf_metadata_as_json(self.metadata_path)
        except Exception as e:
            log_error(f"Failed to load sf-metadata as JSON: {e}")
            self.sf_metadata_string = ""

    @property
    def service_name(self) -> str:
        return get_service_name(self.name, self.version)

    def inject_deployment(self) -> None:
        # JSON 字符串已经是紧凑格式，不需要转义换行符
        json_metadata = self.sf_metadata_string

        envs = {
            "DEEPSEEK_API_KEY": sft_config.deepseek_api_key,
            "DEEPSEEK_BASE_URL": sft_config.deepseek_base_url,
        }

        for env in self.metadata.env:
            envs[env['name']] = env['value']

        env_str = ""
        for key, value in envs.items():
            env_str += f'          - name: {key}\n            value: "{value}"\n'

        # Try to get current git branch and short sha to inject as labels.
        def _sanitize_label_value(v: str, max_len: int = 63) -> str:
            if v is None:
                return ""
            vv = str(v).strip()
            # replace slashes and spaces with '-', keep safe chars
            vv = re.sub(r"[\s/]+", "-", vv)
            vv = re.sub(r"[^0-9A-Za-z_.-]", "", vv)
            vv = vv.lower()
            if len(vv) > max_len:
                vv = vv[:max_len]
            return vv

        git_branch = ""
        git_sha = ""
        try:
            # get branch name
            res = subprocess.run(["git", "rev-parse", "--abbrev-ref", "HEAD"], cwd=self.project_dir, capture_output=True, text=True, check=True)
            git_branch = _sanitize_label_value(res.stdout)
        except Exception:
            git_branch = ""
        try:
            # get short sha
            res2 = subprocess.run(["git", "rev-parse", "--short", "HEAD"], cwd=self.project_dir, capture_output=True, text=True, check=True)
            git_sha = _sanitize_label_value(res2.stdout)
        except Exception:
            git_sha = ""

        deployment_yaml = DEFAULT_DEPLOYMENT_YAML.format(
            registry_address=sft_config.registry_address,
            service_name=self.service_name,
            name=self.name,
            version=self.version,
            namespace=self.namespace,
            sf_metadata=f"'{json_metadata}'",
            env=env_str,
            git_branch=git_branch,
            git_sha=git_sha,
        )
        with open(self.deployment_yaml_path, "w") as f:
            f.write(deployment_yaml)
        print("deployment_yaml_path: ", self.deployment_yaml_path)

    def inject_service_config(self) -> None:
        service_config_path = self.project_dir / Path(self.metadata.service_config)

        config = ServiceConfig.from_yaml_file(service_config_path)

        config.http_port = sft_config.inject_http_port
        config.kafka_host = sft_config.inject_kafka_host
        config.kafka_port = sft_config.inject_kafka_port
        if config.databases is not None:
            for database in config.databases:
                if database.postgres_host is not None:
                    database.postgres_host = sft_config.inject_postgres_host
                    database.postgres_port = sft_config.inject_postgres_port
                    database.postgres_user = sft_config.inject_postgres_user
                    database.postgres_password = sft_config.inject_postgres_password
                    database.postgres_db = self.service_name
                if database.mongo_host is not None:
                    database.mongo_host = sft_config.inject_mongo_host
                    database.mongo_port = sft_config.inject_mongo_port
                    database.mongo_user = sft_config.inject_mongo_user
                    database.mongo_password = sft_config.inject_mongo_password
                    database.mongo_db = sft_config.inject_mongo_db
                if database.redis_host is not None:
                    database.redis_host = sft_config.inject_redis_host
                    database.redis_port = sft_config.inject_redis_port
                    database.redis_password = sft_config.inject_redis_password
        if config.feedback is not None:
            config.feedback.api_url = sft_config.inject_feedback_api_url
            config.feedback.api_timeout = sft_config.inject_feedback_api_timeout
        else:
            config.feedback = ServiceFeedbackConfig(
                api_url=sft_config.inject_feedback_api_url,
                api_timeout=sft_config.inject_feedback_api_timeout,
            )

        if config.signoz is not None:
            config.signoz.api_url = sft_config.inject_signoz_api_url
            config.signoz.api_key = sft_config.inject_signoz_api_key
        else:
            config.signoz = SignozConfig(
                api_url=sft_config.inject_signoz_api_url,
                api_key=sft_config.inject_signoz_api_key,
            )

        if config.trace is not None:
            config.trace.enable = True
            config.trace.url = sft_config.inject_trace_url
            config.trace.headers = sft_config.inject_trace_headers
            config.trace.arg = sft_config.inject_trace_arg
            config.trace.namespace = sft_config.inject_trace_namespace
            config.trace.hostname = sft_config.inject_trace_hostname
        else:
            config.trace = TraceConfig(
                enable=True,
                url=sft_config.inject_trace_url,
                headers=sft_config.inject_trace_headers,
                arg=sft_config.inject_trace_arg,
                namespace=sft_config.inject_trace_namespace,
                hostname=sft_config.inject_trace_hostname,
            )

        if config.entry is not None:
            config.entry.url = sft_config.inject_entry_url
        else:
            config.entry = EntryConfig(
                url=sft_config.inject_entry_url,
            )

        if config.llm is not None:
            config.llm.api_base = sft_config.inject_llm_api_base
            config.llm.api_key = sft_config.inject_llm_api_key
        else:
            config.llm = LLMConfig(
                api_base=sft_config.inject_llm_api_base,
                api_key=sft_config.inject_llm_api_key,
            )

        with open(service_config_path, "w", encoding="utf-8") as f:
            yaml.dump(config.model_dump(), f, allow_unicode=True, indent=2)

    def inject_ingress(self) -> None:
        service_config_path = self.project_dir / Path(self.metadata.service_config)
        config = ServiceConfig.from_yaml_file(service_config_path)

        ingress_yaml = DEFAULT_TRAEFIK_INGRESS_YAML.format(
            name=self.name,
            version=self.version.replace(".", "-"),
            namespace=self.namespace,
            auth_middleware_yaml=DEFAULT_AUTH_MIDDLEWARE_YAML.format(namespace=self.namespace) if config.enable_auth else "",
        )

        for workflow_path in config.workflows:
            workflow_config_path = self.project_dir / Path(self.metadata.service_config).parent / Path(workflow_path)
            if workflow_config_path.exists():
                workflow_config = WorkflowConfig.from_yaml_file(workflow_config_path)
                if config.enable_auth != workflow_config.enable_auth:
                    for node in workflow_config.nodes:
                        if node.type == 'FastAPITrigger' and node.args is not None:
                            path = ""
                            method = ""

                            for key, value in node.args.items():
                                if key == 'path':
                                    path = str(value)
                                if key == 'method':
                                    method = str(value).upper()

                            path = re.sub(r'\{.*?\}', '', path)

                            ingress_yaml += DEFAULT_AUTH_WORKFLOW_YAML.format(
                                name=self.name,
                                version=self.version.replace(".", "-"),
                                namespace=self.namespace,
                                path=path.lstrip('/'),
                                method=method,
                                auth_middleware_yaml=DEFAULT_AUTH_MIDDLEWARE_YAML.format(namespace=self.namespace) if workflow_config.enable_auth else "",
                            )

        with open(self.ingress_yaml_path, "w") as f:
            f.write(ingress_yaml)
        print("ingress_yaml_path: ", self.ingress_yaml_path)

    def inject_dockerfile(self) -> None:
        dockerfile = DEFAULT_DOCKERFILE.format(
            registry_address=sft_config.registry_address,
        )
        with open(self.dockerfile_path, "w") as f:
            f.write(dockerfile)
        print("dockerfile_path: ", self.dockerfile_path)

    def inject_pyproject_toml(self) -> None:
        pyproject_toml = DEFAULT_PYPROJECT_TOML
        with open(self.pyproject_toml_path, "r") as f:
            existing_pyproject_toml = f.read()
        if pyproject_toml.strip() not in existing_pyproject_toml.strip():
            with open(self.pyproject_toml_path, "a") as f:
                f.write(pyproject_toml)
        print("pyproject_toml_path: ", self.pyproject_toml_path)

    def clear_start_sh(self) -> None:
        if Path(self.start_sh_path).exists():
            with open(self.start_sh_path, "rb") as f:
                content = f.read()
            content_str = content.decode("utf-8")
            lines = content_str.splitlines()
            new_content = "\n".join(lines) + ("\n" if content_str.endswith(('\n', '\r')) else "")
            with open(self.start_sh_path, "w", encoding="utf-8", newline="\n") as f:
                f.write(new_content)

    def inject(self) -> None:
        if self.metadata.inject.pyproject_toml:
            self.inject_pyproject_toml()
        if self.metadata.inject.deployment:
            self.inject_deployment()
        if self.metadata.inject.service_config:
            self.inject_service_config()
        if self.metadata.inject.ingress:
            self.inject_ingress()
        if self.metadata.inject.dockerfile:
            self.inject_dockerfile()
        if self.metadata.inject.pyproject_toml:
            self.inject_pyproject_toml()
        self.clear_start_sh()
        save_metadata(self.metadata, self.metadata_path)



from typing import Any
from loguru import logger
from sqlalchemy import select, inspect
from service_forge.workflow.node import Node
from service_forge.workflow.port import Port

class DatabaseCreateOrErrorNode(Node):
    DEFAULT_INPUT_PORTS = [
        Port("data", Any),
        Port("primary_keys", list[str] | None),
    ]

    DEFAULT_OUTPUT_PORTS = [
        Port("data", Any),
    ]

    def __init__(self, name: str):
        super().__init__(name)

    async def _run(self, data: Any, primary_keys: list[str] | None = None) -> None:
        if not hasattr(data, '__table__') or not hasattr(data, '__mapper__'):
            raise ValueError(f"data must be a SQLAlchemy model instance (inherited from declarative_base()), got {type(data)}")
        
        session_factory = await self.default_postgres_database.get_session_factory()
        async with session_factory() as session:
            try:
                model_class = type(data)
                mapper = inspect(model_class)
                
                if primary_keys:
                    pk_columns = []
                    for pk_name in primary_keys:
                        if not hasattr(model_class, pk_name):
                            raise ValueError(f"Field '{pk_name}' does not exist in model {model_class.__name__}")
                        pk_columns.append((pk_name, getattr(model_class, pk_name)))
                else:
                    pk_columns = [(pk_col.name, getattr(model_class, pk_col.name)) for pk_col in mapper.primary_key]
                
                conditions = []
                for pk_name, pk_attr in pk_columns:
                    pk_value = getattr(data, pk_name)
                    if pk_value is None:
                        raise ValueError(f"Primary key '{pk_name}' is None, cannot check existence")
                    conditions.append(pk_attr == pk_value)
                
                query = select(model_class).where(*conditions)
                result = await session.execute(query)
                existing_record = result.scalar_one_or_none()
                
                if existing_record is None:
                    session.add(data)
                    await session.commit()
                    await session.refresh(data)
                    logger.info(f"✓ Created new record in {model_class.__tablename__}")
                    self.activate_output_edges(self.get_output_port_by_name('data'), data)
                else:
                    error_msg = f"Record already exists in {model_class.__tablename__}"
                    raise ValueError(error_msg)
                    
            except Exception as e:
                logger.error(f"✗ Failed to create/retrieve record: {e}")
                await session.rollback()
                raise



# from litellm import completion
# import os

# response = completion(
#     model="openai/ark-deepseek-v3",
#     api_base="http://litellm.vps.shiweinan.com:37919/v1",
#     api_key="sk-2tEpI1fSejYERchVInDU_w",
#     messages=[
#         {"role": "user", "content": "你好，用一句话介绍你自己"}
#     ],
# )

# print(response.choices[0].message.content)

from service_forge.llm import SfLLM, Model
import litellm


# llm = SfLLM(api_base="http://litellm.vps.shiweinan.com:37919/v1", api_key="sk-2tEpI1fSejYERchVInDU_w")
llm = SfLLM(api_base="http://litellm.vps.shiweinan.com:37919/v1", api_key="sk-KVgpceAPL0bg3FXViwuaYw")


response = llm.completion(
    messages=[
        {
            "role": "user",
            "content": [
                {
                    "type": "text",
                    "text": "What’s in this image?"
                },
                {
                    "type": "image_url",
                    "image_url": {
                        "url": "https://awsmp-logos.s3.amazonaws.com/seller-xw5kijmvmzasy/c233c9ade2ccb5491072ae232c814942.png"
                    }
                }
            ]
        }
    ],
    model=Model.QWEN_VL_PLUS,
)

print(response.choices[0].message.content)


from typing import Literal, Optional

from pydantic import BaseModel

class NodeInputPortSchema(BaseModel):
    """
    节点的输入端口定义数据
    """
    name: str
    type: str
    default_value: Optional[str | float | bool | int] = None


class NodeOutputPortSchema(BaseModel):
    """
    节点的输出端口定义数据
    """
    name: str
    type: str


class NodeTypeSchema(BaseModel):
    """
    节点的定义数据
    """
    name: str
    is_trigger: bool
    inputs: dict[str, NodeInputPortSchema]
    outputs: dict[str, NodeOutputPortSchema]



from __future__ import annotations
import uuid
import asyncio
import json
from loguru import logger
from service_forge.workflow.trigger import Trigger
from service_forge.workflow.trigger_event import TriggerEvent
from typing import AsyncIterator, Any
from fastapi import FastAPI, WebSocket, WebSocketDisconnect
from starlette.websockets import WebSocketState
from service_forge.workflow.port import Port
from google.protobuf.message import Message
from google.protobuf.json_format import MessageToJson

class WebSocketAPIV2Trigger(Trigger):
    DEFAULT_INPUT_PORTS = [
        Port("app", FastAPI),
        Port("path", str),
    ]

    DEFAULT_OUTPUT_PORTS = [
        Port("websocket", WebSocket),
        Port("user_id", str),
        Port("token", str),
    ]

    def __init__(self, name: str):
        super().__init__(name)
        self.events = {}
        self.is_setup_websocket = False

    def _setup_websocket(self, app: FastAPI, path: str) -> None:
        async def websocket_handler(websocket: WebSocket):
            websocket.state.user_id = websocket.headers.get("X-User-ID") or "0"
            websocket.state.auth_token = websocket.headers.get("X-User-Token") or ""

            logger.info(f'user_id {websocket.state.user_id} token {websocket.state.auth_token}')
            
            await websocket.accept()

            task_id = uuid.uuid4()
            self.result_queues[task_id] = asyncio.Queue()

            self.trigger_queue.put_nowait({
                "id": task_id,
                "websocket": websocket,
                "user_id": websocket.state.user_id,
                "token": websocket.state.auth_token,
                "data": {},
            })

            await self.result_queues[task_id].get()

        app.websocket(path)(websocket_handler)

    async def _run(self, app: FastAPI, path: str) -> AsyncIterator[bool]:
        if not self.is_setup_websocket:
            self._setup_websocket(app, path)
            self.is_setup_websocket = True

        while True:
            try:
                trigger = await self.trigger_queue.get()
                self.prepare_output_edges('user_id', trigger['user_id'])
                self.prepare_output_edges('token', trigger['token'])
                self.prepare_output_edges('websocket', trigger['websocket'])
                event = TriggerEvent(
                    trigger_node=self,
                    task_id=trigger['id'],
                    trace_context=None,
                    user_id=trigger['user_id'],
                    data=trigger,
                )
                self.trigger(event)
                yield event
            except Exception as e:
                logger.error(f"Error in WebSocketAPITrigger._run: {e}")
                continue

    async def _stop(self) -> AsyncIterator[bool]:
        pass


from service_forge.workflow.node import Node
from service_forge.workflow.nodes.call_service_node import CallServiceNode
from service_forge.workflow.context import Context
from service_forge.sft.config.sf_metadata import load_metadata
from service_forge.workflow.workflow_factory import create_workflow_group
from service_forge.service import Service

def test_get_url():
    metadata = load_metadata("./sf-meta.yaml")
    service = Service.from_config(metadata)

    workflow_config_path = service.config.workflows[0]
    workflow_group = create_workflow_group(
        config_path=service.parse_workflow_path(workflow_config_path),
        entry_config=service.config.entry,
        service_env=service.service_env,
        _handle_stream_output=service._handle_stream_output,
        _handle_query_user=service._handle_query_user,
        database_manager=service.database_manager,
        llm=service.llm,
    )
    main_workflow = workflow_group.get_main_workflow()

    node = main_workflow.nodes[0]
    print(node.get_url(service_name="service-center", version="1.0.0", path="/test", internal=True))
    print(node.get_url(service_name="service-center", version="1.0.0", path="/test", internal=False))
    print(node.get_url(service_name="entry", version="1.0.0", path="/test", internal=True))
    print(node.get_url(service_name="entry", version="1.0.0", path="/test", internal=False))
    print(node.get_url(service_name="data-collection-service", version="1.0.0", path="/test", internal=True))
    print(node.get_url(service_name="data-collection-service", version="1.0.0", path="/test", internal=False))
    print(node.get_url(subdomain="file", service_name="entry", version="1.0.0", path="/test", internal=True))
    print(node.get_url(subdomain="file", service_name="entry", version="1.0.0", path="/test", internal=False))

if __name__ == "__main__":
    test_get_url()


from typing import Type, Union
import types

from fastapi import APIRouter
from omegaconf import OmegaConf
from typing_extensions import get_origin, get_args

from service_forge.model.meta_api.schema import NodeTypeSchema, NodeInputPortSchema, NodeOutputPortSchema
from service_forge.sft.config.sf_metadata import load_metadata
from service_forge.workflow.node import Node, node_register
from service_forge.workflow.trigger import Trigger

def _type_to_string(type_obj: type) -> str:
    """Convert a type to its string representation, handling union types."""
    origin = get_origin(type_obj)
    if origin is Union or origin is types.UnionType:
        args = get_args(type_obj)
        type_names = []
        for arg in args:
            if hasattr(arg, '__name__'):
                type_names.append(arg.__name__)
            else:
                type_names.append(str(arg))
        return ' | '.join(type_names)
    elif hasattr(type_obj, '__name__'):
        return type_obj.__name__
    else:
        return str(type_obj)

def _node_class_to_schema(node_class: Type[Node]) -> NodeTypeSchema:
    """将Node子类转化为NodeTypeSchema"""
    _input_ports = {}
    for port in node_class.DEFAULT_INPUT_PORTS:
        _input_ports[port.name] = NodeInputPortSchema(
            name=port.name,
            type=_type_to_string(port.type),
            default_value=port.default
        )

    _output_ports = {}
    for port in node_class.DEFAULT_OUTPUT_PORTS:
        _output_ports[port.name] = NodeOutputPortSchema(
            name=port.name,
            type=_type_to_string(port.type)
        )

    return NodeTypeSchema(
        name = node_class.__name__,
        inputs=_input_ports,
        outputs=_output_ports,
        is_trigger=issubclass(node_class, Trigger)
    )


def create_schema_router() -> APIRouter:
    router = APIRouter(prefix='/schema')

    @router.get('/nodes', response_model=list[NodeTypeSchema])
    def get_node_schema():
        schemas = []
        for k, v in node_register.items.items():
            if v is not Trigger:
                schemas.append(_node_class_to_schema(v))
        return schemas

    return router

def create_meta_api_router() -> APIRouter:
    router = APIRouter(tags=['meta-api'], prefix='/sdk/meta-api')
    router.include_router(create_schema_router())
    return router

metadata = load_metadata('sf-meta.yaml')
service_config = OmegaConf.load(metadata.service_config)

meta_api_router = create_meta_api_router()



import os

def load_prompt(prompt_path: str) -> str:
    if os.path.exists(prompt_path):
        if prompt_path in prompt_registry:
            return prompt_registry[prompt_path]
        with open(prompt_path, "r") as f:
            prompt = f.read()
            prompt_registry[prompt_path] = prompt
            return prompt
    else:
        return prompt_path

prompt_registry = {}


from typing import Any

from pydantic import BaseModel

from service_forge.workflow.node import Node
from service_forge.workflow.port import Port


class UpdateObjectNode(Node):
    DEFAULT_INPUT_PORTS = [
        Port("object", Any),
        Port("fields", tuple[str, Any], is_extended=True),
    ]

    DEFAULT_OUTPUT_PORTS = [
        Port("object", Any),
    ]

    def __init__(self, name: str) -> None:
        super().__init__(name)

    def _run(
        self,
        object: Any,
        fields: list[tuple[int, tuple[str, Any]]],
    ) -> Any:
        if isinstance(object, BaseModel):
            updates = {f[0]: f[1] for _, f in fields}
            object = object.model_copy(update=updates)
        else:
            for _, f in fields:
                object[f[0]] = f[1]
    
        self.activate_output_edges("object", object)



from __future__ import annotations
import traceback
import asyncio
import uuid
from typing import Any, AsyncIterator, Awaitable, Callable

from loguru import logger
from opentelemetry import trace

from ..db.database import DatabaseManager
from ..trace.trace_scope import workflow_span, node_span, node_output_attributes
from .edge import Edge
from .node import Node
from .port import Port
from .trigger_event import TriggerEvent
from .trigger import Trigger
from ..db.database import DatabaseManager
from .workflow_clone import workflow_clone
from .workflow_callback import WorkflowCallback, BuiltinWorkflowCallback, CallbackEvent
from .workflow_config import WorkflowConfig
from .workflow_validator import WorkflowValidator
from .context import Context
from ..service_config import EntryConfig
from ..llm import SfLLM
from ..sft.config.sf_metadata import metadata
from ..sft.util.name_util import get_url, get_service_name

class Workflow:
    def __init__(
        self,
        id: uuid.UUID,
        config: WorkflowConfig,
        entry_config: EntryConfig,
        nodes: list[Node],
        input_ports: list[Port],
        output_ports: list[Port],
        _handle_stream_output: Callable[[str, AsyncIterator[str]], Awaitable[None]] = None, # deprecated
        _handle_query_user: Callable[[str, str], Awaitable[str]] = None,
        database_manager: DatabaseManager = None,
        llm: SfLLM = None,
        callbacks: list[WorkflowCallback] = [],
        debug_version: bool = False,    # 是否为debug过程中的临时版本

        # for run
        trigger_event: TriggerEvent = None,
        # task_id: uuid.UUID = None,
        real_trigger_node: Trigger = None,

        # global variables
        global_context: Context = None,
    ) -> None:
        self.id = id
        self.config = config
        self.entry_config = entry_config
        self.nodes = nodes
        self.ready_nodes: list[Node] = []
        self.input_ports = input_ports
        self.output_ports = output_ports
        self._handle_stream_output = _handle_stream_output
        self._handle_query_user = _handle_query_user
        self.database_manager = database_manager
        self.llm = llm
        self.run_semaphore = asyncio.Semaphore(config.max_concurrent_runs)
        self.callbacks = callbacks
        self.debug_version = debug_version
        self.trigger_event = trigger_event
        # self.task_id = task_id
        self.real_trigger_node = real_trigger_node
        self.global_context = global_context
        self.progress = 0
    
    @property
    def name(self) -> str:
        return self.config.name
    
    @property
    def version(self) -> str:
        return self.config.version
    
    @property
    def description(self) -> str:
        return self.config.description

    def register_callback(self, callback: WorkflowCallback) -> None:
        self.callbacks.append(callback)

    def unregister_callback(self, callback: WorkflowCallback) -> None:
        self.callbacks.remove(callback)

    async def call_callbacks(self, callback_type: CallbackEvent, *args, **kwargs) -> None:
        for callback in self.callbacks:
            if callback_type == CallbackEvent.ON_WORKFLOW_START:
                await callback.on_workflow_start(*args, **kwargs)
            elif callback_type == CallbackEvent.ON_WORKFLOW_END:
                await callback.on_workflow_end(*args, **kwargs)
            elif callback_type == CallbackEvent.ON_WORKFLOW_ERROR:
                await callback.on_workflow_error(*args, **kwargs)
            elif callback_type == CallbackEvent.ON_NODE_START:
                await callback.on_node_start(*args, **kwargs)
            elif callback_type == CallbackEvent.ON_NODE_END:
                await callback.on_node_end(*args, **kwargs)
            elif callback_type == CallbackEvent.ON_NODE_STREAM_OUTPUT:
                await callback.on_node_stream_output(*args, **kwargs)

    def get_url(
        self,
        service_name: str = "",
        version: str = "",
        path: str = "",
        subdomain: str = "",
        path_params: dict = None,
        internal: bool = None,
    ) -> str:
        if internal is None:
            internal = metadata.mode != "debug"
        if path_params:
            path = path.format(**path_params)
        if internal:
            if subdomain == 'file':
                return f"http://seaweedfs-volume:8080/{path.lstrip('/')}"
            elif service_name == 'entry':
                return f"http://entry-service:80/api/v1/{path.lstrip('/')}"
            elif service_name == 'service-center':
                return f"http://service-center:8080/{path.lstrip('/')}"
            return f"http://{get_service_name(service_name, version)}:80/{path.lstrip('/')}"
        if self.entry_config is None or self.entry_config.url is None:
            logger.error("Entry config is not set, please set the entry config in the service.yaml file.")
            return ""
        return get_url(entry_url=self.entry_config.url, service_name=service_name, version=version, path=path, subdomain=subdomain)

    def add_nodes(self, nodes: list[Node]) -> None:
        for node in nodes:
            node.workflow = self
        self.nodes.extend(nodes)

    def remove_nodes(self, nodes: list[Node]) -> None:
        for node in nodes:
            self.nodes.remove(node)

    def load_config(self) -> None: ...

    def validate(self) -> None:
        validator = WorkflowValidator(self)
        validator.validate()

    def get_input_port_by_name(self, name: str) -> Port:
        for input_port in self.input_ports:
            if input_port.name == name:
                return input_port
        return None

    def get_output_port_by_name(self, name: str) -> Port:
        for output_port in self.output_ports:
            if output_port.name == name:
                return output_port
        return None

    def get_trigger_node(self) -> Trigger:
        trigger_nodes = [node for node in self.nodes if isinstance(node, Trigger)]
        if not trigger_nodes:
            raise ValueError("No trigger nodes found in workflow.")
        if len(trigger_nodes) > 1:
            raise ValueError("Multiple trigger nodes found in workflow.")
        return trigger_nodes[0]

    async def stop(self):
        trigger = self.get_trigger_node()
        await trigger._stop()

    def trigger(self, trigger_name: str, assigned_task_id: uuid.UUID | None, **kwargs) -> uuid.UUID:
        trigger = self.get_trigger_node()
        task_id = assigned_task_id or uuid.uuid4()
        for key, value in kwargs.items():
            trigger.prepare_output_edges(key, value)
        task = asyncio.create_task(self._run(task_id, trigger))
        return task_id

    async def handle_node_stream_output(
        self,
        node: Node,
        stream: AsyncIterator[Any],
    ) -> None:
        async for data in stream:
            await self.call_callbacks(CallbackEvent.ON_NODE_STREAM_OUTPUT, node=node, output=data)

    # TODO: refactor this
    async def handle_query_user(self, node_name: str, prompt: str) -> Awaitable[str]:
        return await asyncio.to_thread(input, f"[{node_name}] {prompt}: ")
    
    def _clone(self, event: TriggerEvent, trigger_node: Trigger) -> Workflow:
        return workflow_clone(self, event, trigger_node)

    async def _run_node_with_callbacks(self, node: Node) -> bool:
        # logger.debug(f"Running node {node.name} with callbacks")
        await self.call_callbacks(CallbackEvent.ON_NODE_START, node=node)
        
        try:
            # logger.debug(f"Node {node.name} running")
            result = node.run()
            # logger.debug(f"Node {node.name} finished running")
            if hasattr(result, '__anext__'):
                await self.handle_node_stream_output(node, result)
            elif asyncio.iscoroutine(result):
                await result
        except Exception as e:
            await self.call_callbacks(CallbackEvent.ON_WORKFLOW_ERROR, workflow=self, node=node, error=e)
            traceback.print_exc()
            logger.error(f"Error when running node {node.name}, task_id: {self.trigger_event.task_id}")
            return False
        finally:
            await self.call_callbacks(CallbackEvent.ON_NODE_END, node=node)

        # logger.debug(f"Node {node.name} callbacks finished")

        return True

    async def _run_node_with_span(self, node: Node) -> bool:
        # logger.debug(f"Running node {node.name} with span")
        async with node_span(
            tracer=trace.get_tracer("service_forge.node"),
            node=node,
        ) as span:
            result = await self._run_node_with_callbacks(node)
            # logger.debug(f"Node {node.name} span finished")
            # node_output_attributes(span, node)
            return result

    async def run_after_trigger(self, event: TriggerEvent) -> Any:
        async with workflow_span(
            tracer=trace.get_tracer("service_forge.workflow"),
            workflow=self,
            parent_context=self.trigger_event.trace_context,
        ):
            logger.info(f"Running workflow: {self.name}")

            await self.call_callbacks(CallbackEvent.ON_WORKFLOW_START, workflow=self)

            self.ready_nodes = []
            trigger_node = self.get_trigger_node()

            for k, v in event.data.items():
                trigger_node.activate_output_edges(k, v)

            for input_port in self.input_ports:
                if input_port.value is not None:
                    input_port.port.node.fill_input(input_port.port, input_port.value)

            for node in self.nodes:
                for key in node.AUTO_FILL_INPUT_PORTS:
                    if key[0] not in [edge.end_port.name for edge in node.input_edges]:
                        node.fill_input_by_name(key[0], key[1])

            running_tasks: dict[asyncio.Task, Node] = {}
            workflow_failed = False
            failed_node = None
            
            while self.ready_nodes or running_tasks:
                logger.info(f"ready_nodes: {self.ready_nodes}, running_tasks: {running_tasks}")
                while self.ready_nodes:
                    node = self.ready_nodes.pop(0)
                    task = asyncio.create_task(self._run_node_with_span(node))
                    running_tasks[task] = node
                
                if not running_tasks:
                    break
                
                done, pending = await asyncio.wait(
                    running_tasks.keys(),
                    return_when=asyncio.FIRST_COMPLETED
                )

                for task in done:
                    node = running_tasks.pop(task)
                    try:
                        result = await task
                        if isinstance(result, Exception):
                            workflow_failed = True
                            failed_node = node
                            break
                        elif result is False:
                            workflow_failed = True
                            failed_node = node
                            logger.error(f"Node {node.name} execution failed, stopping workflow")
                            break
                    except Exception as e:
                        workflow_failed = True
                        failed_node = node
                        logger.error(f"Error when running node {node.name}: {traceback.format_exc()}, task_id: {self.trigger_event.task_id}")
                        break
                
                if workflow_failed:
                    for task in running_tasks.keys():
                        if not task.done():
                            task.cancel()
                    if running_tasks:
                        await asyncio.gather(*running_tasks.keys(), return_exceptions=True)
                    return

            if len(self.output_ports) > 0:
                if len(self.output_ports) == 1:
                    if self.output_ports[0].is_prepared:
                        result = self.output_ports[0].value
                    else:
                        result = None
                else:
                    result = {}
                    for port in self.output_ports:
                        if port.is_prepared:
                            result[port.name] = port.value
                await self.call_callbacks(CallbackEvent.ON_WORKFLOW_END, workflow=self, output=result)
            else:
                await self.call_callbacks(CallbackEvent.ON_WORKFLOW_END, workflow=self, output=None)

    async def _run(self, event: TriggerEvent, trigger_node: Trigger) -> None:
        logger.info(f"run_semaphore状态: 可用={self.run_semaphore._value}, 已使用={self.config.max_concurrent_runs - self.run_semaphore._value}, 最大并发={self.config.max_concurrent_runs}")
        async with self.run_semaphore:
            try:
                new_workflow = self._clone(event, trigger_node)

                result = await new_workflow.run_after_trigger(event)

                if event.task_id in trigger_node.result_queues:
                    trigger_node.result_queues[event.task_id].put_nowait(result)

                for node in new_workflow.nodes:
                    await node.clear()

            except Exception as e:
                traceback.print_exc()
                logger.error(
                    f"Error running workflow: {e}",
                    exc_info=True,
                )
                await self.call_callbacks(
                    CallbackEvent.ON_WORKFLOW_ERROR,
                    workflow=self,
                    node=None,
                    error=e,
                )
                return

    async def run(self):
        tasks = []
        trigger = self.get_trigger_node()

        # event is TriggerEvent
        async for event in trigger.run():
            if isinstance(event, TriggerEvent):
                tasks.append(asyncio.create_task(self._run(event, trigger)))
            else:
                logger.error(f"Unexpected event type: {type(event)}")

        if tasks:
            await asyncio.gather(*tasks)


from __future__ import annotations

from sqlalchemy import Column, String, event
from opentelemetry import trace
from typing import Optional, Any

def get_current_trace_ids() -> tuple[Optional[str], Optional[str]]:
    span = trace.get_current_span()
    if span:
        span_context = span.get_span_context()
        if span_context.is_valid:
            trace_id = format(span_context.trace_id, '032x')
            span_id = format(span_context.span_id, '016x')
            return trace_id, span_id
    return None, None


class TraceMixin:
    trace_id = Column(String(32), nullable=True, comment="OpenTelemetry trace_id")
    span_id = Column(String(16), nullable=True, comment="OpenTelemetry span_id")


def register_trace_events(base: Any) -> None:
    @event.listens_for(base, "before_insert", propagate=True)
    def receive_before_insert(mapper, connection, target):
        if hasattr(target, 'trace_id') and hasattr(target, 'span_id'):
            trace_id, span_id = get_current_trace_ids()
            if trace_id:
                target.trace_id = trace_id
            if span_id:
                target.span_id = span_id

    @event.listens_for(base, "before_update", propagate=True)
    def receive_before_update(mapper, connection, target):
        if hasattr(target, 'trace_id') and hasattr(target, 'span_id'):
            trace_id, span_id = get_current_trace_ids()
            if trace_id:
                target.trace_id = trace_id
            if span_id:
                target.span_id = span_id



from .output.print_node import PrintNode
from .output.kafka_output_node import KafkaOutputNode
from .control.if_node import IfNode
from .control.switch_node import SwitchNode
from .llm.llm_node import LLMNode
from .test.if_console_input_node import IfConsoleInputNode
from .nested.workflow_node import WorkflowNode
from .test.time_consuming_node import TimeConsumingNode

from .db.database_create_or_get_node import DatabaseCreateOrGetNode
from .db.database_create_or_error_node import DatabaseCreateOrErrorNode
from .db.database_update_node import DatabaseUpdateNode
from .db.database_update_fields_node import DatabaseUpdateFieldsNode
from .db.database_get_node import DatabaseGetNode
from .db.database_cast_node import DatabaseCastNode

from .create_object_node import CreateObjectNode
from .update_object_node import UpdateObjectNode
from .call_service_node import CallServiceNode
from .parse_pagination_response_node import ParsePaginationResponseNode
from .parse_list_response_node import ParseListResponseNode

from .any_gate_node import AnyGateNode


from __future__ import annotations
import os
import yaml
from typing import Optional, Any
from pydantic import BaseModel

class SfMetadataInject(BaseModel):
    deployment: bool = True
    service_config: bool = True
    ingress: bool = True
    dockerfile: bool = True
    pyproject_toml: bool = True

class SfMetadataEndpoint(BaseModel):
    path: str
    method: str

class SfMetadataField(BaseModel):
    type: str
    description: str

class SfMetadataParameter(SfMetadataField):
    required: bool = False
    location: str = "body"
    default: Optional[Any] = None
    enum: Optional[list[Any]] = None

class SfMetadataAbility(BaseModel):
    name: str
    description: str
    endpoint: SfMetadataEndpoint
    parameters: dict[str, SfMetadataParameter]
    response: dict[str, SfMetadataField]

class SfMetadata(BaseModel):
    name: str
    version: str
    description: str
    service_config: str
    config_only: bool
    abilities: list[SfMetadataAbility] = []
    env: list[dict]
    inject: SfMetadataInject = SfMetadataInject()
    enable_auth_middleware: bool = True
    mode: str = "debug"

    @classmethod
    def from_yaml_file(cls, filepath: str) -> SfMetadata:
        with open(filepath, 'r', encoding='utf-8') as f:
            data = yaml.safe_load(f)
        return cls(**data)

def load_metadata(path: str) -> SfMetadata:
    if os.path.exists(path):
        return SfMetadata.from_yaml_file(path)
    return None

def save_metadata(meta: SfMetadata, path: str) -> None:
    with open(path, 'w', encoding='utf-8') as f:
        yaml.safe_dump(meta.model_dump(exclude_unset=True, by_alias=True), f, allow_unicode=True)

metadata = load_metadata("sf-meta.yaml")


from __future__ import annotations
from service_forge.workflow.node import Node
from service_forge.workflow.port import Port
from service_forge.workflow.trigger import Trigger
from service_forge.workflow.trigger_event import TriggerEvent
from typing import AsyncIterator
import uuid

class OnceTrigger(Trigger):
    DEFAULT_INPUT_PORTS = [
    ]

    DEFAULT_OUTPUT_PORTS = [
        Port("trigger", bool),
    ]

    def __init__(self, name: str):
        super().__init__(name)

    async def _run(self) -> AsyncIterator[bool]:
        event = TriggerEvent(
            trigger_node=self,
            task_id=uuid.uuid4(),
            trace_context=None,
            user_id="",
            data=None,
        )
        self.trigger(event)
        yield event

    async def _stop(self) -> AsyncIterator[bool]:
        pass


from __future__ import annotations

import asyncio
import json
from abc import abstractmethod

import httpx
from typing import TYPE_CHECKING, Any
from pydantic import BaseModel
from typing_extensions import override
from enum import Enum
from loguru import logger
from .workflow_event import WorkflowResult

if TYPE_CHECKING:
    from .node import Node
    from .workflow import Workflow

class CallbackEvent(Enum):
    ON_WORKFLOW_START = "on_workflow_start"
    ON_WORKFLOW_END = "on_workflow_end"
    ON_WORKFLOW_ERROR = "on_workflow_error"
    ON_NODE_START = "on_node_start"
    ON_NODE_END = "on_node_end"
    ON_NODE_OUTPUT = "on_node_output"
    ON_NODE_STREAM_OUTPUT = "on_node_stream_output"

class WorkflowCallback:
    @abstractmethod
    async def on_workflow_start(self, workflow: Workflow) -> None:
        ...

    @abstractmethod
    async def on_workflow_end(self, workflow: Workflow, output: Any) -> None:
        pass

    @abstractmethod
    async def on_workflow_error(self, workflow: Workflow, node: Node, error: Any) -> None:
        pass

    @abstractmethod
    async def on_node_start(self, node: Node) -> None:
        ...

    @abstractmethod
    async def on_node_end(self, node: Node) -> None:
        ...

    @abstractmethod
    async def on_node_stream_output(self, node: Node, output: Any) -> None:
        ...

class BuiltinWorkflowCallback(WorkflowCallback):
    def __init__(self):
        self._websocket_manager = None
    
    def _get_websocket_manager(self):
        if self._websocket_manager is None:
            from service_forge.api.routers.websocket.websocket_manager import websocket_manager
            self._websocket_manager = websocket_manager
        return self._websocket_manager
    
    def _serialize_result(self, result: Any) -> Any:
        try:
            json.dumps(result)
            return result
        except (TypeError, ValueError):
            return str(result)
    
    def _ensure_data_dict(self, data: Any) -> dict:
        if data is None:
            return {}
        if isinstance(data, dict):
            return data
        if isinstance(data, str):
            try:
                return json.loads(data)
            except (json.JSONDecodeError, TypeError):
                return {"value": data}
        if isinstance(data, BaseModel):
            if hasattr(data, "model_dump"):
                return data.model_dump()
            return data.dict()
        return {"value": data}

    def _get_service_name(self) -> str | None:
        try:
            from service_forge.current_service import get_service
            service = get_service()
            if service is not None:
                return service.config.name
        except Exception:
            pass
        return None

    async def _send_to_websocket(self, workflow: Workflow, message: dict) -> None:
        try: 
            manager = self._get_websocket_manager()
            await manager.send_to_task(workflow.trigger_event.task_id, message)
        except Exception as e:
            logger.error(f"Failed to send message to websocket: {e}")
    
    # TODO: refactor to kafka
    async def _send_service_event(self, workflow: Workflow, event: str, **kwargs) -> None:
        if workflow.entry_config is None or workflow.entry_config.url is None:
            return
        
        async def _do_send() -> None:
            url = None
            try:
                service_name = self._get_service_name()
                if service_name is None:
                    logger.warning("Cannot get service_name, skip sending service-event")
                    return

                url = workflow.get_url(service_name="entry", path="service-event")
                
                payload = {
                    "user_id": str(workflow.trigger_event.user_id),
                    "service_name": service_name,
                    "workflow_name": workflow.name,
                    "task_id": str(workflow.trigger_event.task_id),
                    "event": event,
                    "extra": kwargs
                }

                async with httpx.AsyncClient(timeout=5.0) as client:
                    response = await client.post(
                        url,
                        data=json.dumps(payload, default=str, ensure_ascii=False),
                        headers={"Content-Type": "application/json"},
                    )
                    response.raise_for_status()
            except Exception as e:
                url_str = url if url else workflow.entry_config.url if workflow.entry_config else "unknown"
                logger.error(f"Failed to send service-event to {url_str}: {e}")

        # fire-and-forget: don't await HTTP, just schedule it
        asyncio.create_task(_do_send())
    
    @override
    async def on_workflow_start(self, workflow: Workflow) -> None:
        workflow.progress = 0
        await self._send_service_event(
            workflow,
            "on_workflow_start",
            progress_total=len(workflow.nodes) - 1,
            input=self._ensure_data_dict(workflow.trigger_event.data['data']),
        )

    @override
    async def on_workflow_end(self, workflow: Workflow, output: Any) -> None:
        workflow_result = WorkflowResult(result=output, is_end=True, is_error=False)
        
        if workflow.trigger_event.task_id in workflow.real_trigger_node.result_queues:
            workflow.real_trigger_node.result_queues[workflow.trigger_event.task_id].put_nowait(workflow_result)
        if workflow.trigger_event.task_id in workflow.real_trigger_node.stream_queues:
            workflow.real_trigger_node.stream_queues[workflow.trigger_event.task_id].put_nowait(workflow_result)

        await self._send_service_event(
            workflow,
            "on_workflow_end",
            input=self._ensure_data_dict(workflow.trigger_event.data['data']),
        )

        await self._send_to_websocket(workflow, {
            "type": "workflow_end",
            "task_id": str(workflow.trigger_event.task_id),
            "result": self._serialize_result(output),
            "is_end": True,
            "is_error": False
        })

    @override
    async def on_workflow_error(self, workflow: Workflow, node: Node | None, error: Any) -> None:
        workflow_result = WorkflowResult(result=error, is_end=False, is_error=True)
        
        if workflow.trigger_event.task_id in workflow.real_trigger_node.result_queues:
            workflow.real_trigger_node.result_queues[workflow.trigger_event.task_id].put_nowait(workflow_result)
        if workflow.trigger_event.task_id in workflow.real_trigger_node.stream_queues:
            workflow.real_trigger_node.stream_queues[workflow.trigger_event.task_id].put_nowait(workflow_result)
        
        await self._send_service_event(
            workflow,
            "on_workflow_error",
            input=self._ensure_data_dict(workflow.trigger_event.data['data']),
        )

        await self._send_to_websocket(workflow, {
            "type": "workflow_error",
            "task_id": str(workflow.trigger_event.task_id),
            "node": node.name if node else None,
            "error": self._serialize_result(error),
            "is_end": False,
            "is_error": True
        })

    @override
    async def on_node_start(self, node: Node) -> None:
        await self._send_to_websocket(node.workflow, {
            "type": "node_start",
            "task_id": str(node.workflow.trigger_event.task_id),
            "node": node.name,
            "is_end": False,
            "is_error": False
        })

    @override
    async def on_node_end(self, node: Node) -> None:
        await self._send_to_websocket(node.workflow, {
            "type": "node_end",
            "task_id": str(node.workflow.trigger_event.task_id),
            "node": node.name,
            "is_end": False,
            "is_error": False
        })
        node.workflow.progress += 1
        await self._send_service_event(
            node.workflow,
            "on_node_end",
            progress_total=len(node.workflow.nodes) - 1,
            progress_current=node.workflow.progress,
            input=self._ensure_data_dict(node.workflow.trigger_event.data['data']),
        )

    @override
    async def on_node_stream_output(self, node: Node, output: Any) -> None:
        workflow_result = WorkflowResult(result=output, is_end=False, is_error=False)
        
        if node.workflow.trigger_event.task_id in node.workflow.real_trigger_node.stream_queues:
            node.workflow.real_trigger_node.stream_queues[node.workflow.trigger_event.task_id].put_nowait(workflow_result)

        await self._send_to_websocket(node.workflow, {
            "type": "node_stream_output",
            "task_id": str(node.workflow.trigger_event.task_id),
            "node": node.name,
            "output": self._serialize_result(output),
            "is_end": False,
            "is_error": False
        })
        